;Metroid Improved RNG v1.1 by tertu
;This contains the code for this patch, to be assembled with CA65.
;You assemble this and drop it in place of the save game routines (i.e. $CA35)
;And then edit the rest of the binary.

;What you need to do for the rest of the binary is:
;update the main loop to call random instead of RandomNumbers
;update the code that calls $EE86 to call the modified version of $EE86 here
;replace the beginning of the cartridge ram clear loop with a jump to startup_stuff

;If you're using a disassembly, you can probably just do the following:
;Replace RandomNumbers with random
;Change LEE86 to call random rather than use RandomNumber2
;Add the precalculated RNG seeding stuff to the initialization routines.
;The disassembly that is out there is based on Ophis and this code might not
;actually assemble in Ophis out of the box. I plan to eventually port that to 
;CA65 as well but who knows if that will go anywhere.

;Note that RandomNumber2 won't work anymore. Metroid only uses it in one place,
;but if you use it other places you will have to replace it with calls to
;random.
;Also note that A is not a usable random number after random returns.

;This code is in the public domain, but I would appreciate it if you mentioned
;that this code was used. If I am unable to release this code into the public
;domain, it uses the same license as M. E. O'Neill's sfc.hpp, which is the MIT 
;license.
;Original code: https://gist.github.com/imneme/f1f7821f07cf76504a97f6537c818083

.code

RndA = $9A
RndB = $9B
RndC = $9C
RndD = $9D
RandomNumber1 = $2E

;Returns a random 8-bit number in RandomNumber1
;A port of SCF8 with constants by M. E. O'Neill based on Chris
;Doty-Humphrey's SFC PRNG scheme.
.org $CA35
random:
    ;note: all arithmetic should be unsigned
    ;a, b, c, and d are the input variables
    ;a', b', c', d' are the outputs
    
    ;temp (the result) = a + b + d
    clc
    lda RndD
    adc RndA
    clc
    adc RndB
    sta RandomNumber1
    ;d' = d++
    inc RndD
    ;a' = b eor (b >> 2)
    lda RndB
    lsr a
    lsr a
    eor RndB
    sta RndA
    ;b' = c + (c << 1)
    lda RndC
    asl a
    clc
    adc RndC
    sta RndB
    ;c' = rol8(c, 3) + temp
    lda RndC
    asl 
    adc #0
    asl
    adc #0
    asl
    adc #0
    adc RandomNumber1
    sta RndC
    rts
    
;I have absolutely no idea what this does, but it uses RandomNumber2 and that's
;no good.
LEE86:  lda $B0,x
	bne @end
	txa
	adc $8A
	and #$7F
	sta $B1,x
	pha
	php
	jsr random
	plp
	pla
	adc RandomNumber1
	sta $B2,x
	jsr $EB85
	sta $B3,x
	lda #$01
	sta $B0,x
	rol $8A
@end:   rts

startup_stuff:
;This is the result of a precalculated seeding run with the seed $5EED01.
lda #$EE
sta RndA
lda #$2A
sta RndB
lda #$24
sta RndC
lda #$0D
sta RndD
lda #$B4
sta RandomNumber1

;Handle the instructions our jump destroyed
ldy #$7F
sty $01
jmp $C05B  ;Return to original code.
